# This file is part of Ubuntu Continuous Integration configuration framework.
#
# Copyright 2013, 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

# A tuple containing the five components of the version number: major, minor,
# micro, releaselevel, and serial. All values except releaselevel are integers;
# the release level is 'dev' or 'final'. The
# version_info value corresponding to the uciconfig version 2.0 is (2, 0, 0,
# 'final', 0).
__version__ = (0, 1, 0, 'final', 0)
